#include <stdexcept>
#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include "bindable.h"
#include "buffer.h"
#include "bufferable.h"

using namespace std;

namespace Msp {
namespace GL {

Bufferable::Bufferable():
	buffer(0),
	offset(0),
	next_in_buffer(0),
	prev_in_buffer(0),
	dirty(false)
{ }

Bufferable::~Bufferable()
{
	unlink_from_buffer();
}

void Bufferable::use_buffer(Buffer *buf, Bufferable *prev)
{
	if(prev && buf!=prev->buffer)
		throw invalid_argument("Bufferable::use_buffer");

	if(buffer)
		unlink_from_buffer();

	buffer = buf;
	if(buffer)
	{
		prev_in_buffer = prev;
		if(prev_in_buffer)
		{
			next_in_buffer = prev_in_buffer->next_in_buffer;
			prev_in_buffer->next_in_buffer = this;
		}
	}

	dirty = true;
	update_offset();
}

Bufferable::AsyncUpdater *Bufferable::refresh_async() const
{
	return dirty ? new AsyncUpdater(*this) : 0;
}

void Bufferable::unlink_from_buffer()
{
	if(prev_in_buffer)
		prev_in_buffer->next_in_buffer = next_in_buffer;
	if(next_in_buffer)
	{
		next_in_buffer->prev_in_buffer = prev_in_buffer;
		next_in_buffer->update_offset();
	}
	prev_in_buffer = 0;
	next_in_buffer = 0;
	buffer = 0;
}

void Bufferable::update_offset()
{
	unsigned new_offset = 0;
	if(prev_in_buffer)
		new_offset = prev_in_buffer->offset+prev_in_buffer->get_data_size();

	unsigned align = get_alignment();
	new_offset += align-1;
	new_offset -= new_offset%align;
	if(new_offset!=offset)
	{
		offset = new_offset;
		dirty = true;
		offset_changed();
	}

	if(next_in_buffer)
		next_in_buffer->update_offset();

	/* Do not resize the buffer here, as more bufferables may be added before
	the buffer is actually used. */
}

bool Bufferable::resize_buffer() const
{
	if(offset+get_data_size()>=buffer->get_size())
	{
		const Bufferable *last = this;
		for(; last->next_in_buffer; last=last->next_in_buffer) ;

		unsigned total_size = last->offset+last->get_data_size();

		if(total_size>buffer->get_size())
		{
			buffer->data(total_size, 0);
			return true;
		}
	}

	return false;
}

void Bufferable::update_buffer() const
{
	Conditional<BindRestore> _bind(!ARB_direct_state_access, buffer, buffer->get_type());

	if(resize_buffer())
	{
		/* Resizing the buffer invalidates its contents.  Non-dirty data may
		be in use, so reupload it. */
		for(const Bufferable *b=prev_in_buffer; b; b=b->prev_in_buffer)
			if(!b->dirty)
				b->upload_data(0);
		for(const Bufferable *b=next_in_buffer; b; b=b->next_in_buffer)
			if(!b->dirty)
				b->upload_data(0);
	}

	upload_data(0);
	dirty = false;
}

void Bufferable::upload_data(char *target) const
{
	if(target)
	{
		const char *source = reinterpret_cast<const char *>(get_data_pointer());
		copy(source, source+get_data_size(), target);
	}
	else
		buffer->sub_data(offset, get_data_size(), get_data_pointer());
}


Bufferable::AsyncUpdater::AsyncUpdater(const Bufferable &b):
	bufferable(b)
{
	buffer_resized = bufferable.resize_buffer();
	mapped_address = reinterpret_cast<char *>(bufferable.buffer->map(WRITE_ONLY));
}

Bufferable::AsyncUpdater::~AsyncUpdater()
{
	bufferable.buffer->unmap();
}

void Bufferable::AsyncUpdater::upload_data()
{
	bufferable.upload_data(mapped_address+bufferable.offset);
	// Update all bufferables in the same buffer at once
	for(const Bufferable *b=bufferable.prev_in_buffer; b; b=b->prev_in_buffer)
		if(b->dirty || buffer_resized)
		{
			b->upload_data(mapped_address+b->offset);
			b->dirty = false;
		}
	for(const Bufferable *b=bufferable.next_in_buffer; b; b=b->next_in_buffer)
		if(b->dirty || buffer_resized)
		{
			b->upload_data(mapped_address+b->offset);
			b->dirty = false;
		}
}

} // namespace GL
} // namespace Msp
